class Float
    def round_to(i)
        f = (10 ** i).to_f
        nr = self * f
        return nr.round / f
    end
end

module SkirentalFrontendApi
    class Booking
        include NoBrainer::Document
        include NoBrainer::Document::Timestamps

        before_save :sync_with_backend

        has_many :payments, class_name:SkirentalPayment::Payment, dependent: :destroy

        field :session_id,       :type => String,                       :required => true,   :index => true
        field :shop_id,          :type => String,                       :required => true,   :index => true
        field :context,          :type => Hash,      :default => {}
        field :booking_id,       :type => String
        field :no_sync,          :type => Boolean,   :default => false
        field :changeable_until, :type => String,    :default => nil
        field :creation_time,    :type => String,    :default => nil

        def voucher
            # TODO: Rethink whether this is good for the future. Future: If we want to update bookings or need
            # receipts, we need to store the booking transaction into the payment and have to handle multiple (old) payments
            confirmed_payment = self.payments.order_by(:created_at => :desc).select { |payment|  payment.voucher_code != nil }.last
            if confirmed_payment.present?
                confirmed_payment.voucher_code
            else
                nil
            end
        end

        def last_payment
            self.payments.order_by(:created_at => :desc).last
        end

        def transaction(tax_include = true, tax_show = true, payment_method = nil)
            fail if context.nil?

            current_locale = context[:locale] || "de"
            current_locale = current_locale.split("-").first

            shop_settings = SkirentalApi::Client.settings(shop_id, current_locale)

            items = []
            start_date = Date.parse(context[:startDate]).strftime("%Y-%m-%d")
            duration = (Date.parse(context[:endDate]) - Date.parse(context[:startDate])).to_i + 1

            currency = shop_settings[:money][:payment][:cur]
            price_total = 0
            tax_total = 0

            payment_discount_factor = 1
            unless payment_method.nil?
                payment_providers = shop_settings[:money][:payment][:ipMs]
                payment_providers.each do |provider|
                    if provider[:type] == payment_method
                        payment_discount_factor = (100 - provider[:dsc].to_d) / 100
                        break
                    end
                end
            end

            context[:customers].each do |customer|
                age = customer[:customerGroup][:ageFrom]
                use_insurance = customer[:useInsurance]

                customer[:packages].each do |package|
                    # first, calculate the main product of a package/bundle
                    product_name = package[:name] + (age > 0 ? " [#{customer[:customerGroup][:name]}]" : "")
                    description  = package[:description]
                    item = fetch_price_item(product_name,
                                            description,
                                            package[:mainProduct][:id],
                                            start_date,
                                            duration,
                                            age,
                                            shop_id,
                                            package[:id],
                                            currency,
                                            tax_include,
                                            tax_show,
                                            use_insurance,
                                            current_locale,
                                            package[:mainProduct][:count],
                                            payment_discount_factor)
                    unless item.nil?
                        items.push item
                    end

                    # then add accessories
                    package[:accessories].each do |accessory|
                        product_name = accessory[:name] + (age > 0 ? " [#{customer[:customerGroup][:name]}]" : "")
                        description  = accessory[:description]
                        accessory_item = fetch_price_item(product_name,
                                                          description,
                                                          accessory[:id],
                                                          start_date,
                                                          duration,
                                                          age,
                                                          shop_id,
                                                          package[:id],
                                                          currency,
                                                          tax_include,
                                                          tax_show,
                                                          use_insurance,
                                                          current_locale,
                                                          accessory[:count],
                                                          payment_discount_factor)
                        next if accessory_item.nil?
                        items.push accessory_item
                    end
                end

                # add single products
                customer[:products].each do |product|
                    product_name = product[:name] + (age > 0 ? " [#{customer[:customerGroup][:name]}]" : "")
                    description  = product[:description]
                    product_item = fetch_price_item(product_name,
                                                    description,
                                                    product[:id],
                                                    start_date,
                                                    duration,
                                                    age,
                                                    shop_id,
                                                    nil,
                                                    currency,
                                                    tax_include,
                                                    tax_show,
                                                    use_insurance,
                                                    current_locale,
                                                    product[:count],
                                                    payment_discount_factor)
                    next if product_item.nil?
                    items.push product_item
                end

                self.save
            end

            items.each do |item|
                price_total += item[:price].to_d * item[:quantity]
                if tax_show
                    tax_total += item[:tax].to_d * item[:quantity]
                end
            end


            transaction = {
                :item_list => {
                    :items => items
                },
                :amount =>  {
                    :total    => '%.2f' % ((price_total + tax_total).to_f.round_to(2)),
                    :currency => currency,
                    :details  => !tax_show ? nil : {
                        :subtotal => '%.2f' % price_total.to_f.round_to(2),
                        :tax      => '%.2f' % tax_total.to_f.round_to(2)
                    }
                },
                :description => "Wir bedanken uns fuer Ihre Buchung!" # TODO: do we need a description???
            }

            transaction
        end

        def already_finished
            puts "ALREADY_FINISHED?"
            puts "PAYMENTS_BLANK?: #{self.payments.blank?}"
            puts "CHANGEABLE_UNTIL: #{self.changeable_until}"
            if self.payments.blank?
                return false
            else
                if self.changeable_until.blank? || self.changeable_until == "0001-01-01T00:00:00" # this value is the "zero/undefined" we get from sr-api…
                    puts "FETCHING CURRENT CHANGEABLE_UNTIL_DATE"
                    api_booking = SkirentalApi::Client.get_booking(self.booking_id)
                    api_booking_edit_until = api_booking[:editUntil]
                    puts "API_BOOKING_EDIT_UNTIL: #{api_booking_edit_until}"
                    self.changeable_until = api_booking_edit_until if api_booking_edit_until.present?
                    self.no_sync = true
                    self.save
                end
                reservation_not_changeable = self.changeable_until.to_time < Time.now
                puts "RESERVATION_NOT_CHANGEABLE: #{reservation_not_changeable}"
                self.payments.each do |payment|
                    if payment.is_payed or (payment.is_reserved and reservation_not_changeable)
                        return true
                        break
                    end
                end
                return false
            end
        end

        def sync_with_backend
            puts "SYNC_WITH_BACKEND"
            if self.no_sync
                puts "NO_SYNC: true"
                self.no_sync = false
            else
                puts "NO_SYNC: false"
                puts "PAYMENTS_BLANK?: #{self.payments.blank?}"
                puts "CHANGEABLE_UNTIL: #{self.changeable_until}"
                change_finished_reservation = false # is an already finished RESERVATION allowed to changed (editUntil <= now)?
                if self.payments.present? and self.changeable_until.present?
                    self.payments.each do |payment|
                        if payment.is_reserved and self.changeable_until.to_time >= Time.now
                            change_finished_reservation = true
                            break
                        end
                    end
                end
                if change_finished_reservation
                    puts "CHANGE_FINISHED_RESERVATION -> confirm_booking"
                    api_booking = SkirentalApi::Client.confirm_booking(self, true)
                    puts "API_BOOKING: #{api_booking}"
                    api_booking_id = api_booking[:id]
                    self.booking_id = api_booking_id if api_booking_id.present?
                    api_booking_edit_until = api_booking[:editUntil]
                    self.changeable_until = api_booking_edit_until if api_booking_edit_until.present?
                else
                    puts "CHANGE_UNFINISHED_RESERVATION -> sync_booking_state"
                    api_booking = SkirentalApi::Client.sync_booking_state(self)
                    api_booking_id = api_booking[:id]
                    self.booking_id = api_booking_id if api_booking_id.present?
                    api_booking_edit_until = api_booking[:editUntil]
                    self.changeable_until = api_booking_edit_until if api_booking_edit_until.present?
                end
            end
            true
        end

        def fetch_price_item(name, description, product_id, start_date, duration, age, shop_id, package_id, currency, tax_include, tax_show, use_insurance, current_locale, quantity, discount_factor)
            # 1. We expect that prices are independent to the lang.
            # 2. We store product_id's therefore we do not access prices of packages.
            price = SkirentalApi::Client.get_price_of_item(product_id, start_date, duration, age, shop_id, current_locale, package_id)
            fail if price.nil?
            return nil if price[:price].blank? || price[:price] <= 0 || price[:priceWithInsurance].blank? || price[:priceWithInsurance] <= 0

            price_value = (use_insurance ? price[:priceWithInsurance] : price[:price]) * discount_factor
            tax_value = (use_insurance ? price[:vatValueWithInsurance] : price[:vatValue]) * discount_factor

            # round to 2 decimals because discount_factor could lead to more decimals…
            price_value = '%.2f' % price_value.to_f.round_to(2)
            tax_value = '%.2f' % tax_value.to_f.round_to(2)

            if tax_include and tax_show
                price_value = (price_value.to_d - tax_value.to_d).to_f  # NOTE: paypal wants net prices!
            end

            {
                :name        => name,
                :description => description,
                :quantity    => quantity,
                :price       => price_value,
                :tax         => tax_show ? tax_value : nil,
                :sku         => product_id,
                :currency    => currency
            }
        end
    end
end
